@@ -1,6 +1,6 @@
 #!/usr/bin/perl -w
 
-# $Id: clamdsh.pl,v 1.7 2007/05/18 18:45:11 okamura Exp $
+# $Id: clamdsh.pl,v 1.14 2007/06/16 17:18:34 okamura Exp $
 
 require 5.8.0;
 use strict;
@@ -37,7 +37,7 @@ use vars qw(
 # 定数
 #-------------------------------------------------------------------------------
 # バージョン
-$Version = 'clamdsh.pl v1.0.2';
+$Version = 'clamdsh.pl v1.1.1';
 
 # 終了ステータス
 %ExitStatus = (
@@ -104,7 +104,7 @@ sub Execute {
 			return 1;
 		}
 		else {
-			ClamAV::Shell::UI::Printf("Unknown command: %s\n", $args[0]);
+			printf STDERR "Unknown command: %s\n", $args[0];
 			$ClamAV::Shell::command::Variable{'?'} = -1;
 		}
 		ClamAV::Shell::UI::UpdatePromptVars();
@@ -404,17 +404,26 @@ sub InterpretString {
 
 			if ($subStr =~ m/^(?#
 				)([tnrfbae](?#				TAB や LF など
-				)|1[0-7][0-7](?#			三桁の八進数の文字コード
-				)|[1-7][0-7](?#				二桁の八進数の文字コード
-				)|[0-7](?#					一桁の八進数の文字コード
+				)|[0-7]{1,3}(?#				八進数の文字コード
 				)|x[\da-fA-F]{1,2}(?#		十六進数の文字コード
 				)|x\{[\da-fA-F]{1,4}\}(?#	十六進数のワイド文字コード
-				)|N\{\w+\}|c?(?#			文字名
+				)|N\{[^}]*\}(?#				文字名
+				)|c.(?#						コントロール文字
+				)|l.(?#						一文字の小文字
+				)|u.(?#						一文字の大文字
+				)|L(?:[^\\]|\\[^E])*\E(?#	小文字
+				)|U(?:[^\\]|\\[^E])*\E(?#	大文字
+				)|Q(?:[^\\]|\\[^E])*\E(?#	クォート
 				))/
 			) {
 				my	$token = $1;
+				my	$escaped = "\"\\$token\"";
+				my	$unescaped = eval "
+					use charnames	qw(:full);
+					$escaped;
+				";
 
-				push @interpreted, eval "\"\\$token\"";
+				push @interpreted, $unescaped;
 				$i += length($token) - 1;
 			}
 			else {
@@ -471,11 +480,11 @@ sub InterpretString {
 	unless ($status == 0) {
 		# ダブルクォートが開きっぱなしで終わっている場合
 		if ($status == 1 or $status == 6 or $status == 7) {
-			ClamAV::Shell::UI::Printf("Double quotation is not closed.\n");
+			print STDERR "Double quotation is not closed.\n";
 		}
 		# シングルクォートが開きっぱなしで終わっている場合
 		if ($status == 2 or $status == 8) {
-			ClamAV::Shell::UI::Printf("Single quotation is not closed.\n");
+			print STDERR "Single quotation is not closed.\n";
 		}
 		# エスケープされぱなしで終わっている場合
 		if ($status == 3 or $status == 6 or $status == 8) {
@@ -523,19 +532,17 @@ sub InterpretString {
 sub set {
 	my	@args = @_;
 
-	if ($args[0] =~ m/^(\w+)$/ and scalar(@args) == 2) {
+	if (scalar(@args) == 2 and $args[0] =~ m/^(\w+)$/) {
 		my	($name, $value) = ($args[0], $args[1]);
 
 		if (exists $ClamAV::Shell::Option{$name}) {
 			unless (ClamAV::Shell::Option::SetOption($name, $value)) {
-				ClamAV::Shell::UI::Printf(
-					"Invalid %s value: %s\n", $name, $value
-				);
+				printf STDERR "Invalid %s value: %s\n", $name, $value;
 				return 2;
 			};
 		}
 		elsif ($name eq 'PWD') {
-			ClamAV::Shell::UI::Printf("%s is readonly.", $name);
+			printf STDERR "%s is readonly.\n", $name;
 			return 3;
 		}
 		else {
@@ -543,7 +550,7 @@ sub set {
 		}
 	}
 	else {
-		ClamAV::Shell::UI::Printf("[USAGE]\n%s", $CmdInfo{'set'}->{'usage'});
+		printf STDERR "[USAGE]\n%s\n", $CmdInfo{'set'}->{'usage'};
 		return 1;
 	}
 
@@ -559,12 +566,12 @@ sub set {
 sub unset {
 	my	@args = @_;
 
-	unless ($args[0] =~ m/^\w+$/ and scalar(@args) == 1) {
-		ClamAV::Shell::UI::Printf("[USAGE]\n%s", $CmdInfo{'unset'}->{'usage'});
+	unless (scalar(@args) == 1 and $args[0] =~ m/^\w+$/) {
+		printf STDERR "[USAGE]\n%s\n", $CmdInfo{'unset'}->{'usage'};
 		return 1;
 	}
 	unless ($args[0] ne 'PWD') {
-		ClamAV::Shell::UI::Printf("%s is readonly.", $args[0]);
+		printf STDERR "%s is readonly.\n", $args[0];
 		return 2
 	}
 
@@ -591,7 +598,7 @@ sub unset {
 sub echo {
 	my	@args = @_;
 
-	ClamAV::Shell::UI::Printf("%s", join(' ', @args)); 
+	ClamAV::Shell::UI::Printf("%s\n", join(' ', @args)); 
 
 	return 0;
 }
@@ -606,11 +613,11 @@ sub cd {
 	my	@args = @_;
 
 	unless (scalar(@args) == 1) {
-		ClamAV::Shell::UI::Printf("[USAGE]\n%s", $CmdInfo{'cd'}->{'usage'});
+		printf STDERR "[USAGE]\n%s\n", $CmdInfo{'cd'}->{'usage'};
 		return 1;
 	}
 	unless (chdir($args[0])) {
-		ClamAV::Shell::UI::Printf("%s: %s", "$!", $args[0]);
+		printf STDERR "%s: %s\n", "$!", $args[0];
 		return 2;
 	}
 
@@ -656,12 +663,12 @@ sub help {
 			);
 		}
 		else {
-			ClamAV::Shell::UI::Printf("HELP: unknown topic: %s", $args[0]);
+			printf STDERR "HELP: unknown topic: %s\n", $args[0];
 			return 2;
 		}
 	}
 	else {
-		ClamAV::Shell::UI::Printf("[USAGE]\n%s\n", $CmdInfo{'help'}->{'usage'});
+		printf STDERR "[USAGE]\n%s\n", $CmdInfo{'help'}->{'usage'};
 		return 1;
 	}
 
@@ -681,9 +688,7 @@ sub _SendAndRecv {
 	my	($clamdCmd, @args) = @_;
 
 	unless (scalar(@args) == 0) {
-		ClamAV::Shell::UI::Printf(
-			"[USAGE]\n%s", $CmdInfo{$clamdCmd}->{'usage'}
-		);
+		printf STDERR "[USAGE]\n%s\n", $CmdInfo{$clamdCmd}->{'usage'};
 		return 1;
 	}
 
@@ -699,16 +704,16 @@ sub _SendAndRecv {
 				ClamAV::Shell::UI::Printf("%s", $response);
 			}
 			else {
-				ClamAV::Shell::UI::Printf("%s: %s.", $clamdCmd, "$!");
+				printf STDERR "%s: %s.\n", $clamdCmd, "$!";
 				$ret = 4;
 			}
 		}
 		elsif ($!) {
-			ClamAV::Shell::UI::Printf("%s: %s.", $clamdCmd, "$!");
+			printf STDERR "%s: %s.\n", $clamdCmd, "$!";
 			$ret = 3;
 		}
 		else {
-			ClamAV::Shell::UI::Printf("%s: failed.", $clamdCmd);
+			printf STDERR "%s: failed.\n", $clamdCmd;
 			$ret = 3;
 		}
 		ClamAV::Shell::Clamd::Close($conn);
@@ -781,13 +786,11 @@ sub _SCAN {
 	my	($clamdCmd, @args) = @_;
 
 	unless (scalar(@args) == 1) {
-		ClamAV::Shell::UI::Printf(
-			"[USAGE]\n%s", $CmdInfo{$clamdCmd}->{'usage'}
-		);
+		printf STDERR "[USAGE]\n%s\n", $CmdInfo{$clamdCmd}->{'usage'};
 		return 1;
 	}
 	if ($args[0] =~ m/\n/) {
-		ClamAV::Shell::UI::Printf("%s: include LF :%s", $clamdCmd, $args[0]);
+		printf STDERR "%s: include LF :%s\n", $clamdCmd, $args[0];
 		return 1;
 	}
 
@@ -804,16 +807,16 @@ sub _SCAN {
 				ClamAV::Shell::UI::Printf("%s", $response);
 			}
 			else {
-				ClamAV::Shell::UI::Printf("%s: %s.", $clamdCmd, "$!");
+				printf STDERR "%s: %s.\n", $clamdCmd, "$!";
 				$ret = 4;
 			}
 		}
 		elsif ($!) {
-			ClamAV::Shell::UI::Printf("%s: %s.", $clamdCmd, "$!");
+			printf STDERR "%s: %s.\n", $clamdCmd, "$!";
 			$ret = 3;
 		}
 		else {
-			ClamAV::Shell::UI::Printf("%s: failed.", $clamdCmd);
+			printf STDERR "%s: failed.\n", $clamdCmd;
 			$ret = 3;
 		}
 		ClamAV::Shell::Clamd::Close($conn);
@@ -889,9 +892,7 @@ sub STREAM {
 	my	(@args) = @_;
 
 	unless (scalar(@args) == 1) {
-		ClamAV::Shell::UI::Printf(
-			"[USAGE]\n%s", $CmdInfo{'STREAM'}->{'usage'}
-		);
+		printf STDERR "[USAGE]\n%s\n", $CmdInfo{'STREAM'}->{'usage'};
 		return 1;
 	}
 
@@ -947,9 +948,9 @@ sub STREAM {
 								ClamAV::Shell::UI::Printf("%s", $response);
 							}
 							else {
-								ClamAV::Shell::UI::Printf(
-									"STREAM: %s: %s", "$!", $args[0]
-								);
+								printf STDERR "STREAM: %s: %s\n",
+									"$!", $args[0]
+								;
 								close FH;
 								ClamAV::Shell::Clamd::Close($dconn);
 								$response = ClamAV::Shell::Clamd::Recv($conn);
@@ -957,9 +958,7 @@ sub STREAM {
 							}
 						}
 						else {
-							ClamAV::Shell::UI::Printf(
-								"STREAM: %s: %s", "$!", $args[0]
-							);
+							printf STDERR "STREAM: %s: %s\n", "$!", $args[0];
 							ClamAV::Shell::Clamd::Close($dconn);
 							$response = ClamAV::Shell::Clamd::Recv($conn);
 							ClamAV::Shell::UI::Printf("%s", $response);
@@ -967,30 +966,28 @@ sub STREAM {
 						}
 					}
 					else {
-						ClamAV::Shell::UI::Printf(
-							"STREAM: data port connection failed: %s", $port
-						);
+						printf STDERR
+							"STREAM: data port connection failed: %s\n", $port
+						;
 						$ret = 6;
 					}
 				}
 				else {
-					ClamAV::Shell::UI::Printf(
-						"STREAM: invalid response: %s", $response
-					);
+					printf STDERR "STREAM: invalid response: %s\n", $response;
 					$ret = 5;
 				}
 			}
 			else {
-				ClamAV::Shell::UI::Printf("STREAM: %s.", "$!");
+				printf STDERR "STREAM: %s.\n", "$!";
 				$ret = 4;
 			}
 		}
 		elsif ($!) {
-			ClamAV::Shell::UI::Printf("STREAM: %s.", "$!");
+			printf STDERR "STREAM: %s.\n", "$!";
 			$ret = 3;
 		}
 		else {
-			ClamAV::Shell::UI::Printf("STREAM: failed.");
+			print STDERR "STREAM: failed.\n";
 			$ret = 3;
 		}
 		ClamAV::Shell::Clamd::Close($conn);
@@ -1016,9 +1013,7 @@ sub SESSION {
 	my	(@args) = @_;
 
 	unless (scalar(@args) == 1) {
-		ClamAV::Shell::UI::Printf(
-			"[USAGE]\n%s", $CmdInfo{'SESSION'}->{'usage'}
-		);
+		printf STDERR "[USAGE]\n%s\n", $CmdInfo{'SESSION'}->{'usage'};
 		return 1;
 	}
 
@@ -1037,16 +1032,14 @@ sub SESSION {
 					unless (ClamAV::Shell::Clamd::Send($conn, $line)) {
 						chomp $line;
 						if ($!) {
-							ClamAV::Shell::UI::Printf(
-								"SESSION: send: %s: %s#%d: %s",
+							printf STDERR "SESSION: send: %s: %s#%d: %s\n",
 								"$!", $args[0], $lineNum, $line
-							);
+							;
 						}
 						else {
-							ClamAV::Shell::UI::Printf(
-								"SESSION: send: faild: %s#%d: %s",
+							printf STDERR "SESSION: send: faild: %s#%d: %s\n",
 								$args[0], $lineNum, $line
-							);
+							;
 						}
 						$ret = 5;
 						last;
@@ -1061,10 +1054,9 @@ sub SESSION {
 
 				if ($ret == 0) {
 					unless (ClamAV::Shell::Clamd::Send($conn, "END\n")) {
-						ClamAV::Shell::UI::Printf(
-							"SESSION: send END: %s: %s",
+						printf STDERR "SESSION: send END: %s: %s\n",
 							"$!", $args[0]
-						);
+						;
 						$ret = 6;
 					}
 					while ($response = ClamAV::Shell::Clamd::RecvLine($conn)) {
@@ -1079,18 +1071,16 @@ sub SESSION {
 				}
 			}
 			else {
-				ClamAV::Shell::UI::Printf(
-					"SESSION: open: %s: %s", "$!", $args[0]
-				);
+				printf STDERR "SESSION: open: %s: %s\n", "$!", $args[0];
 				$ret = 4;
 			}
 		}
 		elsif ($!) {
-			ClamAV::Shell::UI::Printf("SESSION: send SESSION: %s.", "$!");
+			printf STDERR "SESSION: send SESSION: %s.\n", "$!";
 			$ret = 3;
 		}
 		else {
-			ClamAV::Shell::UI::Printf("SESSION: send SESSION: failed.");
+			print STDERR "SESSION: send SESSION: failed.\n";
 			$ret = 3;
 		}
 		ClamAV::Shell::Clamd::Close($conn);
@@ -1254,13 +1244,18 @@ $HistoryNumber = 0;
 	# the version of clamdsh.pl
 	qr/\\v/	=> sub {
 		my	$v = $ClamAV::Shell::Version;
-		$v =~ s/clamdsh\.pl v(\d+\.\d+)/$1/;
+		if ($v =~ m/clamdsh\.pl v(\d+\.\d+)/) {
+			$v = "$1";
+		}
+		elsif ($v =~ m/clamdsh\.pl v(\d+)/) {
+			$v = "$1.0";
+		}
 		$v
 	},
 	# the version + patchelvel of clamdsh.pl
 	qr/\\V/	=> sub {
 		my	$v = $ClamAV::Shell::Version;
-		if ($v =~ m/clamdsh\.pl v(\d+\.\d+\.\d+)/) {
+		if ($v =~ m/clamdsh\.pl v(\d+\.\d+(?:\.\d+)+)/) {
 			$v = $1;
 		}
 		elsif ($v =~ m/clamdsh\.pl v(\d+\.\d+)/) {
@@ -1352,7 +1347,7 @@ $HistoryNumber = 0;
 # 初期化
 sub Initialize {
 	unless ($Term) {
-		$Term = new Term::ReadLine 'CLAMD SHell';
+		$Term = new Term::ReadLine 'CLAMD SHell', \*STDIN, \*STDOUT;
 	}
 	UpdatePromptVars();
 }
@@ -1378,7 +1373,14 @@ sub ReadLine {
 
 # 出力
 sub Printf {
-	printf {$Term->OUT} @_;
+	if ($Term and $Term->OUT) {
+		no warnings;
+		printf {$Term->OUT} @_;
+		use warnings;
+	}
+	else {
+		printf STDOUT @_;
+	}
 }
 
 #--- clamd 操作 ----
@@ -1404,14 +1406,13 @@ sub Open {
 			Timeout		=> $ClamAV::Shell::Option{'timeout'},
 		);
 		unless ($conn and $conn->connected()) {
-			ClamAV::Shell::UI::Printf(
-				"%s: %s:%s",
+			printf STDERR "%s: %s:%s\n",
 				"$!",
 				defined($ClamAV::Shell::Option{LocalSocket})
 					? $ClamAV::Shell::Option{'TCPAddr'}
 					: '127.0.0.1',
 				$port
-			);
+			;
 			$conn = undef;
 		}
 	}
@@ -1423,9 +1424,9 @@ sub Open {
 			Peer 		=> $ClamAV::Shell::Option{'LocalSocket'},
 		);
 		unless ($conn and $conn->connected()) {
-			ClamAV::Shell::UI::Printf(
-				"%s: %s", "$!", $ClamAV::Shell::Option{'LocalSocket'}
-			);
+			printf STDERR "%s: %s\n",
+				"$!", $ClamAV::Shell::Option{'LocalSocket'}
+			;
 			$conn = undef;
 		}
 	}
@@ -1443,17 +1444,17 @@ sub Open {
 			Timeout		=> $ClamAV::Shell::Option{'timeout'},
 		);
 		unless ($conn and $conn->connected()) {
-			ClamAV::Shell::UI::Printf(
-				"%s: %s:%s", "$!",
+			printf STDERR "%s: %s:%s",
+				"$!",
 				$ClamAV::Shell::Option{'TCPAddr'},
 				$ClamAV::Shell::Option{'TCPSocket'}
-			);
+			;
 			$conn = undef;
 		}
 	}
 	# ソケットタイプが特定できない場合
 	else {
-		ClamAV::Shell::UI::Printf("Socket type is not specified.");
+		print STDERR "Socket type is not specified.\n";
 		return undef;
 	}
 
@@ -1554,7 +1555,7 @@ sub IsEnableRecv {
 		if ($@ =~ m/^peek timeout\./) {
 			return 0;
 		}
-		ClamAV::Shell::UI::Printf("%s", $@);
+		printf STDERR "%s\n", $@;
 		return undef;
 	}
 
